import helperfun as hf
import numpy as np
from sklearn.decomposition import PCA
import scipy.stats as st


class MoGDecoder:
    def __init__(self, data, label=None, reduce_dim=None, sublabel=None):
        if reduce_dim is not None:
            pca = PCA(n_components=reduce_dim)
            pca.fit(data.transpose())
            data = pca.transform(data.transpose()).transpose()
            
        if label is not None:
            label = hf.mapToZeroN(label)
            self.n_class = max(label) + 1
            if data.ndim == 2:
                # trials are unordered and the class is specified separately
                # data is nDim x nTrial
                self.n_dim = data.shape[0]
                self.n_trial = int(data.shape[1] / self.n_class)  # TODO do something if this is not integer                
                self.data = np.zeros((self.n_class, self.n_dim, self.n_trial))
                if sublabel is not None:
                    self.sublabel = np.zeros((self.n_class, self.n_trial))
                    self.sublabels_per_label = []
                for l in range(self.n_class):
                    self.data[l, :, :] = data[:, label == l]
                    if sublabel is not None:
                        self.sublabel[l, :] = sublabel[label == l]
                        self.sublabels_per_label.append(list(set(sublabel[label == l])))
            else:
                raise ValueError('Data and label dimensions are not specified correctly.')
        else:
            # trials are ordered according to a class dimension
            self.data = data  # nClass x nDim x nTrial
            self.n_class, self.n_dim, self.n_trial = self.data.shape
        self.means = None
        self.vars = None
        self.corrs = None
        self.common_mean = None
        self.common_var = None
        self.common_corr = None
        
    def train(self, train_data):
        n_class, n_dim, n_trial = train_data.shape
        self.means = np.mean(train_data, axis=2)
        self.vars = np.var(train_data, axis=2)
        self.corrs = [np.corrcoef(train_data[c, :, :], rowvar=True) for c in range(n_class)]
        
        all_trials = np.swapaxes(train_data, 1, 2).reshape(n_class * n_trial, n_dim)  # nClass*nTrial x nDim
        self.common_mean = np.mean(all_trials, axis=0)
        self.common_var = np.var(all_trials, axis=0)
        self.common_corr = np.corrcoef(all_trials, rowvar=False)
        
    def evaluate(self, test_data, score='prob_of_truth', mean='CLASS', var='CLASS', corr='CLASS', relative_to_chance=False):
        n_class, n_dim, n_trial = test_data.shape
        likelihoods = np.zeros((n_class, n_trial, n_class))
        for c in range(n_class):
            act_mean = {'CLASS': self.means[c, :], 'COMMON': self.common_mean, 'STANDARD': np.zeros((n_dim))}.get(mean, mean)
            act_var = {'CLASS': self.vars[c, :], 'COMMON': self.common_var, 'STANDARD': np.ones((n_dim))}.get(var, var)
            act_corr = {'CLASS': self.corrs[c], 'COMMON': self.common_corr, 'STANDARD': np.eye(n_dim)}.get(corr, corr)    
            sdmat = np.diag(np.sqrt(act_var))
            act_cov = np.dot(np.dot(sdmat, act_corr), sdmat)
                       
            for true_class in range(n_class):
                for t in range(n_trial):
                    likelihoods[true_class, t, c] = st.multivariate_normal.pdf(test_data[true_class, :, t], mean=act_mean, cov=act_cov)
                        
        class_probs = likelihoods.reshape(n_class * n_trial, n_class)
        class_probs = class_probs / np.tile(np.expand_dims(np.sum(class_probs, axis=1), axis=1), (1, n_class))
        #print(class_probs)
        
        class_labels = np.array([[cl] * n_trial for cl in range(n_class)]).flatten()
        
        if score == 'prob_of_truth':
            scores = class_probs[np.array(range(n_class * n_trial)), class_labels]
        elif score == 'hit_rate':
            #print(np.argmax(class_probs, axis=1))
            scores = np.argmax(class_probs, axis=1) == class_labels
        else:
            raise NotImplementedError('unknown scoring function')
        #print(scores)
        chance = 1.0 / n_class
        if relative_to_chance:
            scores = (scores - chance) / (1 - chance)
        return np.mean(scores)
        
    def crossValidatedPerf(self, n_fold=5, score='prob_of_truth', mean='CLASS', var='CLASS', corr='CLASS', relative_to_chance=False, verbose=False):
        if n_fold == 'by_sublabel':
            n_fold = len(self.sublabels_per_label[0])
            n_test = int(self.n_trial / n_fold)
            folds = []
            for i in range(n_fold):
                act_fold = np.zeros((self.n_class, self.n_dim, n_test))
                for c in range(self.n_class):
                    #print((self.sublabel[c, :].flatten() == self.sublabels_per_label[c][i]).shape)
                    act_fold[c, :, :] = self.data[c, :, self.sublabel[c, :].flatten() == self.sublabels_per_label[c][i]].T
                folds.append(act_fold)
        else:
            n_test = int(self.n_trial / n_fold)
            cv_data = self.data[:, :, :n_test * n_fold]
            folds = [cv_data[:, :, i * n_test:(i + 1) * n_test] for i in range(n_fold)]
        
        avg_scores = np.zeros(n_fold)
        for f in range(n_fold):
            if verbose and f % int(n_fold / 10) == 0:
                print('Fold %d/%d' % (f, n_fold))
            train_data = list(folds)
            del train_data[f]
            train_data = np.concatenate(train_data, axis=2)
            self.train(train_data)
            #print(self.corrs)
            avg_scores[f] = self.evaluate(folds[f], score=score, mean=mean, var=var, corr=corr)
        return avg_scores
