import helperfun as hf
import matplotlib.pyplot as plt
import numpy as np
import pickle
import random
from tensorflow_probability import distributions as tfd
import tensorflow as tf
from decoding import MoGDecoder
import importlib
from sklearn.decomposition import PCA
import scipy as sp
import scipy.stats as st
import skimage
from matplotlib import rcParams


def doubleStatsPlot(categories, axis_labels, fs=16, chance=0, lw=1, maxval=1.1):
    # inspired by Ziemba et al, 2016, PNAS; Fig. 5B
    rcParams['xtick.labelsize'] = fs
    rcParams['ytick.labelsize'] = fs
    # colors = 'rbgcmy'
    colors = ['pink', 'blue', 'green', 'teal', 'olive', 'indigo', 'saddlebrown', 'sienna', 'goldenrod', 'slategray', 'purple']
    
    if len(categories[0]) > len(colors):
        raise NotImplementedError('Plot not implemented for more than %d categories' % len(colors))
    
    # the first category is assumed to be special, and we want to plot the average of the rest
    first = [np.array([]), np.array([])]  
    nonfirst = [np.array([]), np.array([])]
    
    for act_redo in categories:
        for cat_id, actCat in enumerate(act_redo):
            # each category is a 2-element list, and each of these element is an array of values
            # plt.errorbar(np.mean(actCat[0]), np.mean(actCat[1]), xerr=sp.stats.sem(actCat[0]), yerr=sp.stats.sem(actCat[1]), markerfacecolor=colors[cat_id], markeredgecolor=colors[cat_id], marker='o', ecolor='k', capsize=2, markersize=5)
            plt.errorbar(np.mean(actCat[0]), np.mean(actCat[1]), xerr=None, yerr=None, markerfacecolor=colors[cat_id], markeredgecolor=colors[cat_id], marker='o', ecolor='k', capsize=2, markersize=5)
            if cat_id > 0:
                nonfirst[0] = np.concatenate([nonfirst[0], actCat[0]])
                nonfirst[1] = np.concatenate([nonfirst[1], actCat[1]])
            else:
                first[0] = np.concatenate([first[0], actCat[0]])
                first[1] = np.concatenate([first[1], actCat[1]])
    
    plt.errorbar(np.mean(first[0]), np.mean(first[1]), xerr=sp.stats.sem(first[0]), yerr=sp.stats.sem(first[1]), fmt='ro', ecolor='k', capsize=2, markersize=10)
    plt.errorbar(np.mean(nonfirst[0]), np.mean(nonfirst[1]), xerr=sp.stats.sem(nonfirst[0]), yerr=sp.stats.sem(nonfirst[1]), fmt='ko', ecolor='k', capsize=2, markersize=10)
    
    plt.plot([np.mean(nonfirst[0]) - 2, np.mean(nonfirst[0]) + 2],[np.mean(nonfirst[1]) - 2, np.mean(nonfirst[1]) + 2], '--', c='black', linewidth=lw)
    plt.plot([0.0,1.0],[0.0,1.0],'-',c='black',linewidth=lw)
    plt.plot([chance, chance],[0.0, 1],'--',c='black',linewidth=lw)
    plt.plot([0.0, 1],[chance, chance],'--',c='black',linewidth=lw)
    
    plt.xlim([0.8 * chance, maxval])
    plt.ylim([0.8 * chance, maxval])
    plt.xlabel(axis_labels[0],fontsize=fs)
    plt.ylabel(axis_labels[1],fontsize=fs)
    plt.gca().spines['right'].set_visible(False)
    plt.gca().spines['top'].set_visible(False)
    plt.gca().set_aspect(1)


def decoding_test(lvae, arch, textures, labels, samplesPerLabel, dec_dim, trials_per_stim=1000, score='hit_rate', redo=4, plots=True, sampleType='feedforward', noise_amp=0.0):
    nTex = len(set(labels))
    # TODO samplesperlabel can be computed here
    trials_per_fam = trials_per_stim * nTex
    
    all_sample_perfs = [np.array([]) for i in range(lvae.nStochLayer)]
    all_family_perfs = [np.array([]) for i in range(lvae.nStochLayer)]
    #if plots:
    #    f = plt.figure(figsize=(4 * redo, 10))
    all_categories = []
    within = np.zeros(0)
    across = np.zeros(0)
    for r in range(redo):
        if r % 10 == 0:
            print('Resampling %d/%d' % (r, redo))
        idxs = hf.list_slices(range(textures.shape[0]), n_block=nTex, block_length=samplesPerLabel, n_samples=nTex)
        nStim = len(idxs)
        sample_textures = textures[idxs, :]
        sample_texture_families = labels[idxs]
        
        if arch == 'ladder':
            all_inputs = np.vstack([np.tile(sample_textures[s, :], (trials_per_stim, 1)) for s in range(nStim)])
            if noise_amp > 0:
                all_inputs += np.random.normal(loc=0.0, scale=noise_amp, size=all_inputs.shape)
            stim_ids = np.array([[i] * trials_per_stim for i in range(nStim)]).flatten()
            evokedSamples, evokedMeans = lvae.transform(all_inputs, sampleType=sampleType)
            layerNames = ['v%d' % (i + 1) for i in range(lvae.nStochLayer)]
        elif arch == 'stack':
            if not noise_amp == 0.0:
                raise NotImplementedError('noisy decoding not implemented for the stacked architecture')
            evokedSamples, stim_ids = lvae.infer(test_data=sample_textures, nTrials=trials_per_stim, sampleType=sampleType)
            layerNames = lvae.layerNames
            
        corrmats = [np.corrcoef(evokedSamples[0][i * trials_per_stim:(i + 1) * trials_per_stim].T) for i in range(nStim)]
        #corrmats = [np.corrcoef(evokedMeans[0][i * trials_per_stim:(i + 1) * trials_per_stim].T) for i in range(nStim)]
        for s1 in range(nStim):
            for s2 in range(s1 + 1, nStim):
                actDiff = hf.corrMatAbsDiffVals(corrmats[s1], corrmats[s2])
                if sample_texture_families[s1] == sample_texture_families[s2]:
                    within = np.concatenate([within, actDiff])
                else:
                    across = np.concatenate([across, actDiff])
            
        family_ids = hf.mapToZeroN([sample_texture_families[i] for i in stim_ids])
        family_dec_trials = hf.list_slices(range(trials_per_fam * nTex), n_block=nTex, block_length=trials_per_fam, n_samples=trials_per_stim)

        sample_perfs = [[] for i in range(nTex)]
        family_perf = []
        for layer_idx, layerSamples in enumerate(evokedSamples):        
            #family_perf.append(MoGDecoder(data=layerSamples.T[:, family_dec_trials], label=family_ids[family_dec_trials], reduce_dim=dec_dim, sublabel=stim_ids[family_dec_trials]).crossValidatedPerf(n_fold='by_sublabel', score=score, var='STANDARD', corr='STANDARD'))
            family_perf.append(MoGDecoder(data=layerSamples.T, label=family_ids, reduce_dim=dec_dim, sublabel=stim_ids).crossValidatedPerf(n_fold='by_sublabel', score=score, var='STANDARD', corr='STANDARD'))
            #family_perf.append(MoGDecoder(data=layerSamples.T[:, family_dec_trials], label=family_ids[family_dec_trials], reduce_dim=dec_dim).crossValidatedPerf(n_fold=4, score=score, var='STANDARD', corr='STANDARD'))
            for fam in range(nTex):
                sample_perfs[fam].append(MoGDecoder(data=layerSamples.T[:, family_ids == fam], label=stim_ids[family_ids == fam], reduce_dim=dec_dim).crossValidatedPerf(n_fold=4, score=score, var='STANDARD', corr='STANDARD'))
                all_sample_perfs[layer_idx] = np.concatenate([all_sample_perfs[layer_idx], np.array(sample_perfs[fam][-1]).flatten()])
            all_family_perfs[layer_idx] = np.concatenate([all_family_perfs[layer_idx], np.array(family_perf[-1]).flatten()])
            all_categories.append([family_perf] + sample_perfs)
        """if plots:
            rows = 3
            plt.subplot(3, redo, r + 1)
            hf.doubleStatsPlot([family_perf] + sample_perfs, layerNames, chance=1.0 / nTex)
            plt.subplot(3, redo, redo + r + 1)
            plt.boxplot([across, within], labels=['across family', 'within family'])
            plt.subplot(3, redo, 2 * redo + r + 1)
            plt.errorbar([1, 2], [np.mean(np.array(cond)) for cond in [across, within]], 
                         yerr=[st.sem(np.array(cond)) for cond in [across, within]])"""
    if plots:
        plt.figure(figsize=(14, 4))
        plt.subplot(1, 3, 1)
        doubleStatsPlot(all_categories, layerNames, chance=1.0 / nTex)
        plt.subplot(1, 3, 2)
        plt.boxplot([across, within], labels=['across family', 'within family'])
        plt.subplot(1, 3, 3)
        plt.errorbar([1, 2], [np.mean(np.array(cond)) for cond in [across, within]], 
                         yerr=[st.sem(np.array(cond)) for cond in [across, within]])
    return all_sample_perfs, all_family_perfs


def get_datasets(npx=16, nTex=3, plot=False, label_preference=[6, 2, 0, 3, 4, 5, 1]): 
    filtered = hf.loadImages(28, 80000, kind="filtered", folder='./')
    filtered = hf.cropImages(filtered, npx)
    if plot:
        hf.image_set(filtered[:10, :], unitscale=False)

    texture_data = hf.loadImages(28, 42000, kind="textures", use_labels=label_preference[:nTex], folder='./')
    samplesPerLabel = 5000 # ???? int(texture_data['train_images'].shape[0] / nTex)
    textures = hf.cropImages(texture_data['train_images'], npx)
    labels = texture_data['train_labels']

    # normalizing each class separately for luminance and contrast
    for f in set(labels):
        act_text = textures[labels == f, :]
        act_text = act_text - np.tile(np.mean(act_text, axis=0), (act_text.shape[0], 1))
        act_text = act_text / np.tile(np.std(act_text, axis=0), (act_text.shape[0], 1))
        textures[labels == f] = act_text
    if plot:
        hf.image_set(np.vstack([textures[i * samplesPerLabel:i*samplesPerLabel + int(20 / nTex), :] for i in range(nTex)]), unitscale=False)
    return filtered, textures, labels, samplesPerLabel


def normalized_mnist(npx=None):
    (textures, labels), _ = tf.keras.datasets.mnist.load_data()
    if npx is not None:
        temp = np.zeros((textures.shape[0], npx, npx))
        for img in range(textures.shape[0]):
            temp[img, :, :] = skimage.transform.resize(textures[img, :, :], (npx, npx))
        textures = temp
    textures = textures.reshape((textures.shape[0], -1))
    std_image = np.std(textures, axis=0)
    std_image[std_image == 0] = 1
    textures = textures - np.tile(np.mean(textures, axis=0), (textures.shape[0], 1))
    textures = textures / np.tile(std_image, (textures.shape[0], 1))
    return textures, labels


def plot_rec_rank(sam, rec):
    plt.figure(figsize=(20, 3))
    n_rec = 5
    for i in range(n_rec):
        idx = np.argsort(sam[i, :])
        plt.subplot(2, n_rec, i + 1)
        plt.bar(range(sam.shape[1]), sam[i, idx])
        plt.gca().set_xticks([])
        plt.gca().set_yticks([])
        plt.subplot(2, n_rec, n_rec + i + 1)
        plt.bar(range(sam.shape[1]), rec[i, idx])
        plt.gca().set_xticks([])
        plt.gca().set_yticks([])
